import os
import shutil
from collections.abc import ItemsView, KeysView, ValuesView
from itertools import chain
from pathlib import Path
from typing import IO, Any, Dict, Union
from urllib.parse import urlparse

from astropy.coordinates import EarthLocation
from astropy.units import Quantity
from tomlkit.exceptions import ParseError

from . import environ
from .data_type import RichParameters, ValueRange
from .exceptions import NECSTConfigurationError, NECSTParameterNameError
from .files import read
from .inform import get_logger

DefaultNECSTRoot = Path.home() / ".necst"
DefaultsPath = Path(__file__).parent / ".." / "defaults"
logger = get_logger(__name__)


parsers = dict(
    location=lambda x: EarthLocation(**x),
    antenna_drive_range_az=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_range_el=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_warning_limit_az=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_warning_limit_el=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_critical_limit_az=lambda x: ValueRange(*map(Quantity, x)),
    antenna_drive_critical_limit_el=lambda x: ValueRange(*map(Quantity, x)),
    antenna_pointing_parameter_path=Path,
)


class Configuration(RichParameters):
    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except NECSTParameterNameError as e:
            # Translate to specific error
            raise NECSTConfigurationError from e

    @classmethod
    def from_file(cls, __path: Union[os.PathLike, str, IO]):
        try:
            return super().from_file(__path)
        except ParseError as e:
            # Translate to specific error
            raise NECSTConfigurationError from e

    @property
    def _dotnecst(self) -> str:
        from urllib.parse import urlparse

        try:
            url = urlparse(self._metadata.get("path", ""))
            scheme = f"{url.scheme}://" if url.scheme else ""
            path = url.path
            path = path.decode("utf-8") if isinstance(path, bytes) else path
            return scheme + url.netloc + str(Path(path).parent)
        except Exception:
            return ""

    def reload(self):
        """Reload the configuration.

        This method is to reflect the changes made to the configuration file or the
        change of environment variables.

        """
        path = find_config()
        new = self.__class__.from_file(path)

        slots = chain.from_iterable(
            getattr(cls, "__slots__", []) for cls in self.__class__.__mro__
        )
        for attr in slots:
            setattr(self, attr, getattr(new, attr))

        for key, parser in parsers.items():
            try:
                self.attach_parsers(**{key: parser})
            except NECSTParameterNameError:
                pass

    @classmethod
    def configure(cls):
        """Create config file under default directory ``$HOME/.necst``."""
        DefaultNECSTRoot.mkdir(exist_ok=True)
        for file in DefaultsPath.glob("*.toml"):
            _target_path = DefaultNECSTRoot / file.name
            if _target_path.exists():
                logger.error(f"'{_target_path}' already exists, skipping...")
                continue
            shutil.copyfile(file, _target_path)
        cls().reload()

    @property
    def parameters(self) -> Dict[str, Any]:
        unwrapped = {}
        for k, v in self._parameters.items():
            parsed = v.parsed
            if isinstance(parsed, Path):
                if urlparse(self._dotnecst).scheme:
                    unwrapped[k] = os.path.join(self._dotnecst, parsed)
                else:
                    unwrapped[k] = Path(self._dotnecst) / parsed
            else:
                unwrapped[k] = parsed
        return unwrapped

    def keys(self) -> KeysView:
        return self.parameters.keys()

    def values(self) -> ValuesView:
        return self.parameters.values()

    def items(self) -> ItemsView:
        return self.parameters.items()

    def __getitem__(self, key: str, /):
        item = super().__getitem__(key)
        if isinstance(item, Path):
            if urlparse(self._dotnecst).scheme:
                item = os.path.join(self._dotnecst, item)
            else:
                item = Path(self._dotnecst) / item
        return item

    def __getattr__(self, key: str, /):
        attr = super().__getattr__(key)
        if isinstance(attr, Path):
            if urlparse(self._dotnecst).scheme:
                attr = os.path.join(self._dotnecst, attr)
            else:
                attr = Path(self._dotnecst) / attr
        return attr

    def __add__(self, other: Any, /):
        if other is None:
            return self
        if not isinstance(other, self.__class__):
            return NotImplemented

        if self._prefix == other._prefix:
            parameters = {k: v.parsed for k, v in self._parameters.items()}
            for k, v in other._parameters.items():
                if (k in parameters) and (parameters[k] != v):
                    raise KeyError(
                        f"Conflicting values for {k!r} found: {parameters[k]} and {v}."
                    )
                parameters[k] = v
            parameters.update(other._parameters)
            inst = self.__class__(self._prefix, **parameters)
            return inst

        length = len(self._prefix) + 1
        prefix_stripped = {
            k[length:]: v.parsed
            for k, v in self._parameters.items()
            if k.startswith(self._prefix)
        }
        length = len(other._prefix) + 1
        prefix_stripped_other = {
            k[length:]: v.parsed
            for k, v in other._parameters.items()
            if k.startswith(other._prefix)
        }
        for k, v in prefix_stripped_other.items():
            if (k in prefix_stripped) and (prefix_stripped[k] != v):
                raise KeyError(
                    f"Conflicting values for {k!r} found:"
                    f" {prefix_stripped[k]!r} and {v!r}."
                )
            prefix_stripped[k] = v
        return self.__class__(**prefix_stripped)

    __radd__ = __add__


def find_config() -> str:
    """Look for the configuration file from environment variables and defaults."""
    root_candidates = [str(DefaultNECSTRoot)]
    specified = environ.necst_root.get()
    if specified is not None:
        root_candidates.insert(0, specified)
    config_file_name = "config.toml"
    candidates = map(lambda x: os.path.join(x, config_file_name), root_candidates)

    for path in candidates:
        try:
            read(path)
            return path
        except FileNotFoundError:
            logger.debug(f"Config file not found at {path!r}")

    logger.error(
        "Config file not found, using the default parameters. "
        "To create the file with default parameters, run `neclib.configure()`."
    )
    config_path = str(DefaultsPath / "config.toml")
    return str(config_path)


config = Configuration()
config.reload()

configure = Configuration.configure



from typing import Callable, Dict, List

import astropy.units as u

from ...core import logic
from ...utils import sanity_check
from .ad_converter_base import ADConverter


class CPZ3177(ADConverter):

    """a/d converter, which can convert by 32 or 64 channels.

    Notes
    -----
    Configuration items for this device:

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default RSW1
        setting of 0. This ID would be non-zero, when multiple PCI board of same model
        are mounted on a single FA (Factory Automation) controller.

    ave_num : int
        Number of sampled data to be averaged over. This script not only measures the
        actual voltage, but also performs fluctuation reduction in order to measure a
        faint voltage. The value should not be too large compared to sampling frequency
        ``smpl_freq``, otherwise the output won't follow the change of voltage.

    smpl_freq : int
        Sampling frequency; number of data obtained in 1 second. The value should not
        be too small against ``ave_num``.

    single_diff : {"SINGLE" or "DIFF"}
        Input type of voltage. "SINGLE" is Single-ended input, "DIFF" is
        "Differential input".

    all_ch_num : int
        Number of channels used in voltage measurement. This number must be the same
        as the number of channels defined for ``smpl_ch_req``. The maximum number is
        32 in differential input, 64 in single-ended input. Please read the manual
        of this board for wiring design.

    smpl_ch_req : List[Dict[str, Union[int, str]]]
        List of measurement range, in format
        ``{ ch_no = int(channel_number), range = str(range_specifier) }``. The length of
        the list must be the same as ``all_ch_num``. In addition, it must be defined
        from ch1 through the maximum channel number to be used. For example, when
        channels [3, 5, 9, 12] are used, ``smpl_ch_req`` must be set from ch1 through
        ch12. The acceptable "range" values are as below:
        '0_5V': 0 - 5 V
        '010V' : 0 - 10 V
        '2P5V' : -2.5 - 2.5 V
        '5V' : -5 - 5 V
        '10v' : -10 - 10 V
        These should be set to the same value as the value corresponding to combination
        of three DIP switch "DSW1", "DSW2", "DSW3" mounted on the side of the board.
        Please read the manual of this board for the combination of DIP switches.

    channel : Dict[str, int]
        Human-readable channel name. The value should be mapping from human readable
        version (str) to device level identifier (int). You can assign any name to the
        channels. No need to define the aliases for all the channels listed in
        ``smpl_ch_req``, but defining aliases for unused channels will raise error.

    converter : Dict[str, str]
        Functions to convert measured voltage to any parameter you want, in format
        ``{str(parameter_type) = str(function)}``. Supported ``parameter_types`` are
        ["V", "I", "P"], and ``x`` in ``function`` will be substituted by the measured
        value. This would be useful when measured voltage is scaled and/or shifted
        version of phisical parameter.

    """

    Manufacturer = "Interface"
    Model = "CPZ3177"

    Identifier = "rsw_id"

    def __init__(self, **kwargs) -> None:
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.ave_num = self.Config.ave_num
        self.smpl_freq = self.Config.smpl_freq
        self.single_diff = self.Config.single_diff
        self.all_ch_num = self.Config.all_ch_num
        self.smpl_ch_req = self.Config.smpl_ch_req

        self.ad = pyinterface.open(3177, self.rsw_id)
        self.ad.stop_sampling()
        self.ad.initialize()
        self.ad.set_sampling_config(
            smpl_ch_req=self.smpl_ch_req,  # It must be a list
            smpl_num=1000,
            smpl_freq=self.smpl_freq,
            single_diff=self.single_diff,
            trig_mode="ETERNITY",
        )
        self.ad.start_sampling("ASYNC")

    def get_data(self, ch: int) -> List[float]:
        with logic.busy(self, "busy"):
            offset = self.ad.get_status()["smpl_count"] - self.ave_num
            data = self.ad.read_sampling_buffer(self.ave_num, offset)
            data_li_2 = []
            for i in range(self.all_ch_num):
                data_li = []
                for k in range(self.ave_num):
                    data_li.append(data[k][i])
                data_li_2.append(data_li)

            ave_data_li = []
            for data in data_li_2:
                d = sum(data) / self.ave_num
                ave_data_li.append(d)
            return ave_data_li[ch - 1]

    @property
    def converter(self) -> Dict[str, Callable[[float], float]]:
        _ = [sanity_check(expr, "x") for expr in self.Config.converter.values()]
        return {k: eval(f"lambda x: {v}") for k, v in self.Config.converter.items()}

    def get_voltage(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        return self.converter["V"](self.get_data(ch)) * u.mV

    def get_current(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        return self.converter["I"](self.get_data(ch)) * u.microampere

    def get_power(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        return self.converter["P"](self.get_data(ch)) * u.mW

    def finalize(self) -> None:
        self.ad.stop_sampling()



__all__ = ["clip"]

from typing import Optional, TypeVar, overload

from ..type_aliases import SupportsComparison

T = TypeVar("T", bound=SupportsComparison)


@overload
def clip(value: T, minimum: T, maximum: T) -> T:
    ...


@overload
def clip(value: T, *, absmax: T) -> T:
    ...


def clip(
    value: T,
    minimum: Optional[T] = None,
    maximum: Optional[T] = None,
    *,
    absmax: Optional[T] = None,
) -> T:
    """Limit the ``value`` to the range [``minimum``, ``maximum``].

    Parameters
    ----------
    value
        Arbitrary parameter.
    minimum
        Lower limit of ``value``.
    maximum
        Upper limit of ``value``.
    absmax
        Upper limit of absolute value of ``value``.

    Examples
    --------
    >>> neclib.utils.clip(1.2, 0, 1)
    1
    >>> neclib.utils.clip(41, 0, 100)
    41
    >>> neclib.utils.clip(-4, absmax=3)
    -3

    """
    if absmax is not None:
        maximum = abs(absmax)  # type: ignore
        minimum = -1 * maximum  # type: ignore
    if minimum > maximum:  # type: ignore
        raise ValueError("Minimum should be less than maximum.")
    return min(max(minimum, value), maximum)  # type: ignore



"""Collection of data type definitions this package interprets."""

from .ordinal import Ordinal
from .parameters import Parameters
from .rich_parameters import RichParameters
from .value_range import ValueRange

__all__ = ["Ordinal", "Parameters", "RichParameters", "ValueRange"]



from __future__ import annotations

from io import StringIO
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

from ..inform import get_logger
from ..type_aliases import CoordFrameType

if TYPE_CHECKING:
    # Circular import
    from ...coordinates.observations.observation_spec_base import ObservationSpec

__all__ = ["html_repr_of_observation_spec"]

logger = get_logger(__name__)


def html_repr_of_observation_spec(
    observation_spec: ObservationSpec, /, *, frame: CoordFrameType = "fk5"
) -> str:
    """Return HTML representation of the observation specification."""

    waypoints = iter(observation_spec)
    waypoint_repr = []

    _svg = StringIO()
    with plt.style.context("dark_background"), plt.rc_context(
        {"font.family": "serif", "font.size": 9}
    ):
        fig, ax = plt.subplots(figsize=(3, 3))
        ax.set(
            xlabel="Longitude [deg]",
            ylabel="Latitude [deg]",
            title=f"Waypoints in {frame.upper()} frame",
            aspect=1,
        )
        if frame.lower() not in ["altaz", "horizontal", "azel"]:
            ax.invert_xaxis()  # Celestial coordinate frames are generally right-handed.
        pt_kw = dict(ms=5, alpha=0.9)
        last_coord = None
        for i, wp in enumerate(waypoints):
            if i > 100:  # Limit the number of waypoints to be shown.
                waypoint_repr.append('<tr><td colspan="3">...</td></tr>')
                break

            coord = wp.coordinates.transform_to(frame)
            lon, lat = coord.data.lon.to_value("deg"), coord.data.lat.to_value("deg")
            try:
                nan_coord = (lon != lon) or (lat != lat)
            except ValueError:
                nan_coord = (lon != lon).any() or (lat != lat).any()

            if (coord.size == 0) or nan_coord:
                pass  # Cannot determine where to plot.
            elif coord.size == 1:
                if last_coord is not None:
                    _lon = [last_coord[0], lon]
                    _lat = [last_coord[1], lat]
                    ax.plot(_lon, _lat, c=wp.mode.__class__.DRIVE.value, zorder=-1)
                ax.plot(lon, lat, ".", c=wp.mode.value, **pt_kw)
                last_coord = (lon, lat)
            else:
                if last_coord is not None:
                    _lon = [last_coord[0], lon[0]]
                    _lat = [last_coord[1], lat[0]]
                    ax.plot(
                        _lon, _lat, c=wp.mode.__class__.DRIVE.value, zorder=-1, lw=1
                    )
                ax.plot(lon, lat, c=wp.mode.value, lw=0.3)
                last_coord = (lon[-1], lat[-1])

            try:
                _coord_str = [
                    f"({_lon:9.4f}, {_lat:9.4f})" for _lon, _lat in zip(lon, lat)
                ]
                coord_str = " &#x2192; ".join(_coord_str)
            except TypeError:
                coord_str = f"({lon:9.4f}, {lat:9.4f})"
            others = []
            for info in ["speed", "scan_frame", "integration", "id"]:
                attr = getattr(wp, info)
                if attr is not None:
                    others.append(f"{info}={attr}")
            waypoint_repr.append(
                f"""
                <tr>
                    <td>{wp.mode.name}</td>
                    <td><code>{coord_str}</code></td>
                    <td>{", ".join(others)}</td>
                </tr>
                """
            )
        ax.grid(True, c="#757")
        fig.tight_layout()
        fig.savefig(_svg, format="svg")
    _svg.seek(0)
    svg = _svg.read()
    plt.close(fig)  # Don't show the figure in Jupyter Notebook inline.

    return f"""
    <details><summary>{len(waypoint_repr)} waypoints</summary>
        <table>
            <thead>
                <tr>
                    <th>Mode</th>
                    <th>{frame.upper()} Coordinate [deg]</th>
                    <th>Other Parameters</th>
                </tr>
            </thead>
            <tbody>
                {''.join(waypoint_repr)}
            </tbody>
        </table>
        <span>The time interval for OFF and HOT observations are not considered.</span>
    </details>
    <details open><summary>Graph</summary>
        {svg}
    </details>
    """



import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Generator, Iterator, Optional, Tuple, Union

import astropy.units as u
import matplotlib.pyplot as plt
import pandas as pd
from astropy.coordinates import SkyCoord

from ...core import Parameters
from ...core.formatting import html_repr_of_observation_spec
from ...core.type_aliases import CoordFrameType, CoordinateType, UnitType
from ...core.units import scan_to_points


class ObservationMode(Enum):
    """Type of observation to be performed at certain coordinate."""

    DRIVE = "#777"
    ON = "#0F5"
    OFF = "#0DF"
    HOT = "#F50"
    SKY = "#0DF"


class TimeKeeper:
    """Judge whether it's time to run constant interval observation or not."""

    def __init__(self, interval: u.Quantity, points_per_scan: int = 1):
        self.interval = interval
        self.last = None
        if self._time_based:
            self.count = 0
        else:
            self.count = u.Quantity(0, self.interval.unit)
        self.points_per_scan = points_per_scan

    @property
    def should_observe(self) -> bool:
        """Return ``True`` if it's time to run observation, otherwise ``False``."""
        if self._time_based:
            self.count = time.time() * u.s
        if self.last is None:
            return True
        return bool((self.count - self.last) > self.interval)

    @property
    def _time_based(self) -> bool:
        return self.interval.unit.is_equivalent(u.s)

    def increment(self, unit: UnitType) -> None:
        """Increment the counter, which will be compared with ``interval``."""
        if not self._time_based:
            _unity = u.Quantity(1, unit or self.interval.unit)
            unity = _unity.to(
                self.interval.unit, equivalencies=scan_to_points(self.points_per_scan)
            )
            self.count += unity

    def tell_observed(self) -> None:
        """Tell the time keeper that the observation has been completed."""
        self.last = time.time() * u.s if self._time_based else self.count


@dataclass
class Waypoint:
    mode: ObservationMode
    """Observation mode."""
    target: Optional[Union[str, CoordinateType]] = None
    """Coordinate or name of the target object."""
    reference: Optional[Union[str, CoordinateType]] = None
    """Coordinate or name of the object the other coordinate is defined relative to."""
    scan_frame: Optional[CoordFrameType] = None
    """Coordinate frame in which the scan is performed."""
    start: Optional[Tuple[u.Quantity, u.Quantity]] = None
    """Start position of the scan. The frame is defined in ``scan_frame``."""
    stop: Optional[Tuple[u.Quantity, u.Quantity]] = None
    """Stop position of the scan. The frame is defined in ``scan_frame``."""
    offset: Optional[CoordinateType] = None
    """Offset applied to all the coordinates specified above."""
    speed: Optional[u.Quantity] = None
    """Scan speed."""
    integration: Optional[u.Quantity] = None
    """Integration time."""
    id: Any = None
    """Identifier of the coordinate, if necessary."""

    @property
    def is_scan(self) -> bool:
        """Whether this coordinate contains enough information to perform a scan."""
        params = (self.start, self.stop, self.speed, self.scan_frame)
        return all(x is not None for x in params)

    @property
    def with_offset(self) -> bool:
        """Whether this coordinate should be offset by certain amount."""
        return self.offset is not None

    @property
    def name_query(self) -> bool:
        """Whether the target is specified by its name."""
        return isinstance(self.target, str) or isinstance(self.reference, str)

    @property
    def coordinates(self) -> SkyCoord:
        """Return a list of coordinates this section represents.

        Warning
        -------
        This property doesn't perform accurate coordinate calculation when coordinates
        in AltAz frame is involved. Please use :mod:`neclib.coordinates.convert` module
        for accurate coordinate calculation.

        """
        import numpy as np
        from astropy.time import Time

        from ...core import config
        from ..convert import CoordCalculator
        from ..frame import parse_frame

        if not hasattr(self, "_calc"):
            self._calc = CoordCalculator(config.location)
        now = Time(time.time(), format="unix")

        nowhere = (float("nan") * u.deg, float("nan") * u.deg, "fk5")

        if self.name_query:
            coord = self._calc.get_body(self.target or self.reference, now)
        else:
            target = self.target or self.reference or nowhere
            coord = self._calc.get_skycoord(*target[:2], frame=target[2], obstime=now)

        if self.with_offset:
            offset_frame = self.offset[2]
            if isinstance(offset_frame, str):
                offset_frame = parse_frame(offset_frame)
            converted = coord.transform_to(offset_frame)
            lon = converted.data.lon + self.offset[0]
            lat = converted.data.lat + self.offset[1]
            coord = self._calc.get_skycoord(lon, lat, frame=self.offset[2], obstime=now)

        if self.is_scan:
            now_broadcasted = np.broadcast_to(now, (2,))
            scan_frame = self.scan_frame
            if isinstance(scan_frame, str):
                scan_frame = parse_frame(scan_frame)
            converted = coord.transform_to(scan_frame)
            ref_lon, ref_lat = converted.data.lon, converted.data.lat
            lon = np.r_[ref_lon + self.start[0], ref_lon + self.stop[0]]
            lat = np.r_[ref_lat + self.start[1], ref_lat + self.stop[1]]
            coord = self._calc.get_skycoord(
                lon, lat, frame=self.scan_frame, obstime=now_broadcasted
            )

        return coord


class ObservationSpec(Parameters, ABC):
    __slots__ = ("_executing", "_coords", "_fig")
    _repr_frame: CoordFrameType = "fk5"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._executing: Optional[Generator[Waypoint, None, None]] = None
        self._coords: Optional[Waypoint] = None
        self._fig: Optional[plt.Figure] = None

    @property
    def coords(self) -> pd.DataFrame:
        """Crudely calculated waypoints this object represents.

        Warning
        -------
        This property doesn't perform accurate coordinate calculation when coordinates
        in AltAz frame is involved.

        """
        if self._coords is None:
            self._coords = ...
            raise NotImplementedError
        return self._coords

    @property
    def fig(self) -> plt.Figure:
        """Figure of crudely calculated telescope driving path.

        Notes
        -----
        If you need ``Axes`` object, use ``fig.axes`` attribute.

        """
        if self._fig is None:
            self._fig = ...
            raise NotImplementedError
        return self._fig

    @abstractmethod
    def observe(self) -> Generator[Waypoint, None, None]:
        """Define all the observation steps which will be interpreted and commanded."""
        ...

    def __next__(self) -> Waypoint:
        if self._executing is None:
            self._executing = self.observe()

        # If generator created above contain nothing, raise StopIteration
        if self._executing is None:
            raise StopIteration
        return next(self._executing)

    def __iter__(self) -> Iterator[Waypoint]:
        return self

    def _repr_html_(self) -> str:
        # TODO: Use properties ``coords`` and ``fig`` to show the path
        return super()._repr_html_() + html_repr_of_observation_spec(
            self, frame=self._repr_frame
        )

    def hot(self, id: Any = None, integration: Optional[u.Quantity] = None):
        # Infer the duration by general keyword
        if integration is None:
            integration = self["integ_hot"]
        return Waypoint(mode=ObservationMode.HOT, id=id, integration=integration)

    def off(
        self,
        id: Any = None,
        integration: Optional[u.Quantity] = None,
        *,
        relative: Optional[bool] = None,
        on_coord: Optional[CoordinateType] = None,
        off_coord: Optional[CoordinateType] = None
    ) -> Waypoint:
        # Infer the duration by general keyword
        if integration is None:
            integration = self["integ_off"]
        if relative is None:
            relative = self["relative"]
        if on_coord is None:
            on_coord = (self["lambda_on"], self["beta_on"], self["coord_sys"])
        if off_coord is None:
            if relative:
                lon, lat = self["delta_lambda"], self["delta_beta"]
                off_coord = (lon, lat, self["coord_sys"])
            else:
                off_coord = (self["lambda_off"], self["beta_off"], self["coord_sys"])

        kwargs = dict(mode=ObservationMode.OFF, id=id, integration=integration)
        if relative:
            kwargs.update(reference=on_coord, offset=off_coord)
        else:
            kwargs.update(target=off_coord)
        return Waypoint(**kwargs)



from typing import Optional, Union

import astropy.units as u

from ..type_aliases import DimensionLess, UnitType


def get_quantity(
    *value: Union[DimensionLess, u.Quantity], unit: Optional[UnitType] = None
) -> u.Quantity:
    """Convert a value to astropy Quantity.

    Parameters
    ----------
    value
        The value to be converted. If multiple values are given, they should have
        equivalent dimensions.
    unit
        The unit of the value. Should be given as a keyword argument.

    Returns
    -------
    quantity
        The value converted to Quantity.

    Raises
    ------
    UnitConversionError
        If the value has incompatible dimensions.
    ValueError
        If the values have different shapes.

    Examples
    --------
    >>> neclib.core.get_quantity(1, 2, 3, unit="m")
    <Quantity [1., 2., 3.] m>
    >>> neclib.core.get_quantity(np.array([1, 2, 3]), unit="m/s")
    <Quantity [1., 2., 3.] m / s>

    """
    if len(value) == 1:
        # Don't make 1-D array when there's only one value
        return u.Quantity(value[0], unit=unit)
    return u.Quantity(value, unit=unit)



__all__ = ["PathFinder", "standby_position", "CoordinateGeneratorManager"]

import math
import time as pytime
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import astropy.units as u
import numpy as np
from astropy.coordinates.name_resolve import NameResolveError

from .. import config, utils
from ..core.type_aliases import CoordFrameType, DimensionLess, UnitType
from .convert import CoordCalculator

T = TypeVar("T", DimensionLess, u.Quantity)


def standby_position(
    start: Tuple[T, T],
    end: Tuple[T, T],
    *,
    margin: Union[float, int, u.Quantity],
    unit: Optional[UnitType] = None,
) -> Tuple[u.Quantity, u.Quantity]:
    """Calculate the standby position taking into account the margin during scanning.

    Parameters
    ----------
    start
        Longitude and latitude of start point.
    end
        Longitude and latitude of end point.
    unit
        Angular unit in which longitude and latitude are given. If they are given as
        ``Quantity``, this parameter will be ignored.

    Examples
    --------
    >>> neclib.coordinates.standby_position(start=(30, 50), end=(32, 50), unit="deg")
    (<Quantity 29. deg>, <Quantity 50. deg>)

    """
    margin = utils.get_quantity(margin, unit=unit)
    start = utils.get_quantity(*start, unit=unit)
    end = utils.get_quantity(*end, unit=unit)
    return CoordCalculator.get_extended(start, end, length=margin)


def get_position_angle(start_lon, end_lon, start_lat, end_lat):
    return CoordCalculator.get_position_angle(
        (start_lon, end_lon), (start_lat, end_lat)
    )


class Timer:
    def __init__(self):
        self.start = pytime.time()
        self.target = self.start

    def set_offset(self, offset) -> None:
        if offset != float(offset):
            raise TypeError(f"Cannot parse offset {offset} as float value")
        self.target += float(offset)

    def get(self) -> float:
        return self.target

    def __bool__(self) -> bool:
        return True


@dataclass
class ControlStatus:
    controlled: bool = True
    tight: Optional[bool] = None
    """If True, the section is for observation so its accuracy should be inspected."""
    start: Optional[float] = None
    """Start time of this control section."""
    stop: Optional[float] = None
    """End time of this control section."""
    infinite: bool = False
    """Whether the control section is infinite hence need interruption or not."""
    waypoint: bool = False
    """Whether this is waypoint hence need some value to be sent or not."""

    def __bool__(self) -> bool:
        return True


CoordinateGenerator = Generator[
    Tuple[u.Quantity, u.Quantity, List[float], ControlStatus], Literal[True], None
]


class PathFinder(CoordCalculator):
    """望遠鏡の軌道を計算する

    Parameters
    ----------
    location
        Location of observatory.
    pointing_param_path
        Path to pointing parameter file.
    pressure
        Atmospheric pressure at the observation environment.
    temperature
        Temperature at the observation environment.
    relative_humidity
        Relative humidity at the observation environment.
    obswl
        Observing wavelength.

    Attributes
    ----------
    location: EarthLocation
        Location of observatory.
    pressure: Quantity
        Atmospheric pressure, to compute diffraction correction.
    temperature: Quantity
        Temperature, to compute diffraction correction.
    relative_humidity: Quantity or float
        Relative humidity, to compute diffraction correction.
    obswl: Quantity
        Observing wavelength, to compute diffraction correction.
    obsfreq: Quantity
        Observing frequency of EM-wave, to compute diffraction correction.

    Examples
    --------
    >>> location = EarthLocation("138.472153deg", "35.940874deg", "1386m")
    >>> path = "path/to/pointing_param.toml"
    >>> pressure = 850 * u.hPa
    >>> temperature = 290 * u.K
    >>> humid = 0.5
    >>> obsfreq = 230 * u.GHz
    >>> finder = neclib.coordinates.PathFinder(
    ...     location, path, pressure=pressure, temperature=temperature,
    ...     relative_humidity=humid, obsfreq=obsfreq)

    """

    command_unit_duration_sec = 1

    @property
    def unit_n_cmd(self) -> int:
        """Number of commands to be calculated in ``self.command_unit_duration_sec``."""
        return int(self.command_unit_duration_sec * config.antenna_command_frequency)

    @property
    def unit_index(self) -> List[float]:
        return self.index()

    def index(self, n_cmd: Optional[Union[float, int]] = None) -> List[float]:
        if n_cmd is None:
            n_cmd = self.unit_n_cmd
        idx = [i / n_cmd for i in range(math.floor(n_cmd))]
        if n_cmd != math.floor(n_cmd):
            idx.append(n_cmd)
        return idx

    def time_index(
        self, start_time: float, n_cmd: Optional[Union[float, int]] = None
    ) -> List[float]:
        if n_cmd is None:
            n_cmd = self.unit_n_cmd
        start_time += config.antenna_command_offset_sec
        command_freq = config.antenna_command_frequency
        time = [start_time + i / command_freq for i in range(math.floor(n_cmd))]
        if n_cmd != math.floor(n_cmd):
            time.append(start_time + n_cmd / (command_freq - 1))
        return time

    def functional(
        self,
        lon: Callable[[float], T],
        lat: Callable[[float], T],
        frame: CoordFrameType,
        *,
        unit: Optional[UnitType] = None,
        n_cmd: Union[int, float],
        mode: ControlStatus,
        time: Optional[Timer] = None,
    ) -> CoordinateGenerator:
        """Calculate the path of antenna movement from any function.

        Parameters
        ----------
        lon
            Function that returns longitude. The function should accept single argument;
            auxiliary variable of value between 0 and 1.
        lat
            Function that returns latitude.
        frame
            Frame in which the coordinates are given.
        unit
            Angular unit in which longitude and latitude are given. If they are given as
            ``Quantity``, this parameter will be ignored.
        n_cmd
            DimensionLess of commands. The path will be calculated supplying arithmetic
            sequence of auxiliary variable, which first term is 0, last is 1, number of
            members is ``n_cmd``.
        time
            Start time manager. If ``None``, current time will be used. No consideration
            for ``antenna_command_offset_sec`` is required.

        Returns
        -------
        Az, El, Time, ControlStatus
            Tuple of conversion result, azimuth, elevation, time and type information of
            the control section.

        Examples
        --------
        >>> def lon(x):
        ...     return 30 + x * 2
        >>> def lat(x):
        ...     return 50
        >>> path = finder.functional(lon, lat, frame="altaz", unit="deg", n_cmd=1000)
        >>> next(path)
        (<Quantity [30., 30.1, 30.2, ...] deg>, <Quantity [50., 50., 50., ...] deg>,
        [1610612736.0, 1610612736.1, 1610612736.2, ...])

        """
        time = time or Timer()
        start = time.get() + config.antenna_command_offset_sec
        time.set_offset(n_cmd / config.antenna_command_frequency)

        mode.start = start
        mode.stop = time.get() + config.antenna_command_offset_sec

        unit_n_cmd = self.unit_n_cmd

        for seq in range(math.ceil(n_cmd / unit_n_cmd)):
            idx = [seq * unit_n_cmd + i for i in range(unit_n_cmd)]
            param = [_idx / n_cmd for _idx in idx if _idx <= n_cmd]
            idx = idx[: len(param)]
            if (seq == math.ceil(n_cmd / unit_n_cmd) - 1) and (param[-1] < 1):
                param.append(1)
                idx.append(idx[-1] + (param[-1] - param[-2]) * n_cmd)

            t_for_this_seq = [start + i / config.antenna_command_frequency for i in idx]
            lon_for_this_seq = [lon(p) for p in param]
            lat_for_this_seq = [lat(p) for p in param]

            yield *self.get_altaz(
                lon=lon_for_this_seq,
                lat=lat_for_this_seq,
                frame=frame,
                unit=unit,
                obstime=t_for_this_seq,
            ), mode

    def standby(
        self,
        lon: Callable[[float], float],
        lat: Callable[[float], float],
        frame: CoordFrameType,
        *,
        at: Union[int, float] = 0,
        unit: Optional[UnitType] = None,
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        mode = mode or ControlStatus()
        if not mode.waypoint:
            mode.infinite = True
        while True:
            yield from self.functional(
                lambda x: lon(at),
                lambda x: lat(at),
                frame=frame,
                unit=unit,
                n_cmd=self.unit_n_cmd,
                time=time,
                mode=mode,
            )

    def accelerate(
        self,
        lon: Callable[[float], T],
        lat: Callable[[float], T],
        frame: CoordFrameType,
        *,
        length: Union[int, float],
        margin: Union[int, float, u.Quantity],
        speed: Union[float, int, u.Quantity],
        unit: UnitType,
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.infinite = False
        mode.tight = False

        margin = utils.get_quantity(margin, unit=unit)
        length = utils.get_quantity(length, unit=unit)
        speed = abs(utils.get_quantity(speed, unit=margin.unit / u.s))

        self.logger.warning(
            "Calculation involving AltAz coordinate may contain a jump of speed "
            "(~0.001deg/s) on transition from acceleration mode to linear drive, due to"
            " insufficient consideration of sidereal motion"
        )

        start_param = -margin / length
        scale = -1 / start_param

        if margin.value == 0:
            required_time = 0 * u.s
        else:
            # マージン部分の座標計算 加速度その1
            a = (speed**2) / (2 * margin)

            # マージン部分の座標計算 加速度その2
            # a_az = config.antenna_max_acceleration_az
            # a_el = config.antenna_max_acceleration_el
            # a = (a_az**2 + a_el**2) ** (1 / 2)
            required_time = ((2 * margin) / a) ** (1 / 2)
        n_cmd = required_time.to_value("s") * config.antenna_command_frequency

        def scaled(x):
            propto_accel = x**2
            return propto_accel / scale + start_param

        tracking_ok = None
        mode.waypoint = True
        for result in self.standby(
            lon, lat, frame, at=start_param, time=time, mode=mode
        ):
            tracking_ok = yield result
            if tracking_ok is not None:
                break
        mode.waypoint = False

        # Duration this drive will take (required_time) isn't taken into account, so the
        # functions `lon` and `lat` can be different from the ones used in linear path
        # calculation that will follow this. The angular error is approximately equal to
        # (sidereal motion speed * required_time).
        yield from self.functional(
            lambda x: lon(scaled(x)),
            lambda x: lat(scaled(x)),
            frame=frame,
            n_cmd=n_cmd,
            time=time,
            mode=mode,
        )

    def linear(
        self,
        start: Tuple[T, T],
        end: Tuple[T, T],
        frame: CoordFrameType,
        *,
        speed: Union[float, int, u.Quantity],
        unit: Optional[UnitType] = None,
        margin: Union[int, float, u.Quantity],
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        """望遠鏡の直線軌道を計算する

        Parameters
        ----------
        start
            Longitude and latitude of start point.
        end
            Longitude and latitude of end point.
        frame
            Coordinate frame, in which longitude and latitude are given.
        speed
            Telescope drive speed.
        unit
            Angular unit in which longitude and latitude are given. If they are given as
            ``Quantity``, this parameter will be ignored.

        Returns
        -------
        Az, El, Time
            Tuple of calculated azimuth, elevation and time commands.

        Examples
        --------
        >>> finder.linear(
            start=(0, 0), end=(0.05, 0), frame="altaz", speed=0.5, unit=u.deg
        )
        [<Quantity [-1.47920569, -1.46920569, -1.45920569, -1.44920569, -1.43920569,
           -1.42920569] deg>, <Quantity [-1.88176239, -1.88176188, -1.88176136,
           -1.88176084, -1.88176032, -1.8817598 ] deg>, array([1.66685637e+09,
           1.66685637e+09, 1.66685637e+09, 1.66685637e+09,
           1.66685637e+09, 1.66685637e+09])]

        """
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.infinite = False

        start = utils.get_quantity(*start, unit=unit)
        end = utils.get_quantity(*end, unit=unit)
        speed = abs(utils.get_quantity(speed, unit=f"{end.unit}/s"))
        distance = ((start - end) ** 2).sum() ** 0.5
        n_cmd = (distance / speed) * (config.antenna_command_frequency * u.Hz)

        def lon(x):
            return start[0] + x * (end[0] - start[0])

        def lat(x):
            return start[1] + x * (end[1] - start[1])

        yield from self.accelerate(
            lon,
            lat,
            frame,
            length=distance,
            margin=margin,
            speed=speed,
            unit=end.unit,
            time=time,
            mode=ControlStatus(tight=False),
        )

        mode.tight = True
        yield from self.functional(
            lon, lat, frame, unit=unit, n_cmd=n_cmd, mode=mode, time=time
        )

    def offset_linear(
        self,
        start: Tuple[T, T],
        end: Tuple[T, T],
        frame: CoordFrameType,
        *,
        reference: Tuple[T, T, CoordFrameType],
        speed: Union[float, int, u.Quantity],
        unit: Optional[UnitType] = None,
        margin: Union[int, float, u.Quantity],
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.infinite = False

        start = utils.get_quantity(*start, unit=unit)
        end = utils.get_quantity(*end, unit=unit)
        speed = abs(utils.get_quantity(speed, unit=f"{end.unit}/s"))
        distance = ((start - end) ** 2).sum() ** 0.5
        n_cmd = (distance / speed) * (config.antenna_command_frequency * u.Hz)

        t = self.time_index(time.get(), n_cmd)
        idx = self.index(n_cmd)
        ref = self.get_skycoord(
            *reference[:2], frame=reference[2], obstime=t, unit=unit
        )
        ref = self.transform_to(ref, frame)

        def lon(x):
            ref_lon = np.interp(x, idx, ref.data.lon)
            return ref_lon + start[0] + x * (end[0] - start[0])

        def lat(x):
            ref_lat = np.interp(x, idx, ref.data.lat)
            return ref_lat + start[1] + x * (end[1] - start[1])

        yield from self.accelerate(
            lon,
            lat,
            frame,
            length=distance,
            margin=margin,
            speed=speed,
            unit=end.unit,
            time=time,
            mode=ControlStatus(tight=False),
        )

        mode.tight = True
        yield from self.functional(
            lon, lat, frame, unit=unit, n_cmd=n_cmd, time=time, mode=mode
        )

    def offset_linear_by_name(
        self,
        start: Tuple[T, T],
        end: Tuple[T, T],
        frame: CoordFrameType,
        *,
        name: str,
        speed: Union[float, int, u.Quantity],
        unit: Optional[UnitType] = None,
        margin: Union[int, float, u.Quantity],
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.infinite = False

        start = utils.get_quantity(*start, unit=unit)
        end = utils.get_quantity(*end, unit=unit)
        speed = abs(utils.get_quantity(speed, unit=f"{end.unit}/s"))
        distance = ((start - end) ** 2).sum() ** 0.5
        n_cmd = (distance / speed) * (config.antenna_command_frequency * u.Hz)

        t = self.time_index(time.get(), n_cmd)
        idx = self.index(n_cmd)
        if frame == "altaz":
            frame = self._get_altaz_frame(t)
        try:
            ref = self.get_body(name, t)
            ref = self.transform_to(ref, frame)
        except NameResolveError:
            self.logger.error(f"Cannot resolve {name!r}")
            return f"Cannot resolve {name!r}"

        def lon(x):
            ref_lon = ref.data.lon
            ref_lon = np.interp(x, idx, ref_lon) if ref_lon.size > 1 else ref_lon
            return ref_lon + start[0] + x * (end[0] - start[0])

        def lat(x):
            ref_lat = ref.data.lat
            ref_lat = np.interp(x, idx, ref_lat) if ref_lat.size > 1 else ref_lat
            return ref_lat + start[1] + x * (end[1] - start[1])

        yield from self.accelerate(
            lon,
            lat,
            frame,
            length=distance,
            margin=margin,
            speed=speed,
            unit=end.unit,
            time=time,
            mode=ControlStatus(tight=False),
        )

        mode.tight = True
        yield from self.functional(lon, lat, frame, n_cmd=n_cmd, time=time, mode=mode)

    def track(
        self,
        lon: T,
        lat: T,
        frame: CoordFrameType,
        *,
        unit: Optional[UnitType] = None,
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.tight = True
        if not mode.waypoint:
            mode.infinite = True
        while True:
            yield from self.functional(
                lambda x: lon,
                lambda x: lat,
                frame,
                unit=unit,
                n_cmd=self.unit_n_cmd,
                time=time,
                mode=mode,
            )

    def track_by_name(
        self,
        name: str,
        *,
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.tight = True
        if not mode.waypoint:
            mode.infinite = True

        def lon(x):
            coord_data = coord.data.lon
            if coord_data.size == 1:
                return coord_data
            return np.interp(x, self.unit_index, coord_data)

        def lat(x):
            coord_data = coord.data.lat
            if coord_data.size == 1:
                return coord_data
            return np.interp(x, self.unit_index, coord_data)

        while True:
            start_time = time.get()
            t = self.time_index(start_time)
            try:
                coord = self.get_body(name, t)
            except NameResolveError:
                self.logger.error(f"Cannot resolve {name!r}")
                return f"Cannot resolve {name!r}"

            yield from self.functional(
                lon, lat, coord.frame.name, n_cmd=self.unit_n_cmd, time=time, mode=mode
            )

    def track_with_offset(
        self,
        lon: T,
        lat: T,
        frame: CoordFrameType,
        offset: Tuple[T, T, CoordFrameType],
        *,
        unit: Optional[UnitType] = None,
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.tight = True
        if not mode.waypoint:
            mode.infinite = True
        offset_lon, offset_lat = utils.get_quantity(*offset[:2], unit=unit)

        def f_lon(x):
            return np.interp(x, self.unit_index, offset_applied_lon)

        def f_lat(x):
            return np.interp(x, self.unit_index, offset_applied_lat)

        while True:
            t = self.time_index(time.get())
            ref = self.get_skycoord(lon, lat, frame=frame, obstime=t, unit=unit)
            ref_in_target_frame = self.transform_to(ref, offset[2])
            offset_applied_lon = ref_in_target_frame.data.lon + offset_lon
            offset_applied_lat = ref_in_target_frame.data.lat + offset_lat
            yield from self.functional(
                f_lon,
                f_lat,
                offset[2],
                unit=unit,
                n_cmd=self.unit_n_cmd,
                time=time,
                mode=mode,
            )

    def track_by_name_with_offset(
        self,
        name: str,
        offset: Tuple[T, T, CoordFrameType],
        *,
        unit: Optional[UnitType] = None,
        time: Optional[Timer] = None,
        mode: Optional[ControlStatus] = None,
    ) -> CoordinateGenerator:
        time = time or Timer()
        mode = mode or ControlStatus()
        mode.tight = True
        if not mode.waypoint:
            mode.infinite = True
        offset_lon, offset_lat = utils.get_quantity(*offset[:2], unit=unit)

        def lon(x):
            return np.interp(x, self.unit_index, offset_applied_lon)

        def lat(x):
            return np.interp(x, self.unit_index, offset_applied_lat)

        while True:
            start_time = time.get()
            t = self.time_index(start_time)
            try:
                ref = self.get_body(name, t)
            except NameResolveError:
                self.logger.error(f"Cannot resolve {name!r}")
                return f"Cannot resolve {name!r}"

            frame = offset[2] if offset[2] != "altaz" else self._get_altaz_frame(t)
            ref_in_target_frame = self.transform_to(ref, frame)
            offset_applied_lon = ref_in_target_frame.data.lon + offset_lon
            offset_applied_lat = ref_in_target_frame.data.lat + offset_lat

            yield from self.functional(
                lon, lat, offset[2], n_cmd=self.unit_n_cmd, time=time, mode=mode
            )


class CoordinateGeneratorManager:
    def __init__(self, generator: Optional[CoordinateGenerator] = None) -> None:
        self._generator = generator
        self._send_value = None

    def will_send(self, value: Any) -> None:
        self._send_value = value

    def __iter__(self) -> Iterable[Any]:
        return self

    def __next__(self) -> Any:
        if self._generator is None:
            self._send_value = None
            raise StopIteration("No generator attached")
        if self._send_value is None:
            return next(self._generator)
        try:
            ret = self._generator.send(self._send_value)
            self._send_value = None
            return ret
        except TypeError:
            # Keep send value once for just-started generator
            return next(self._generator)

    def attach(self, generator: CoordinateGenerator) -> None:
        self.clear()
        self._generator = generator

    def clear(self) -> None:
        if self._generator is not None:
            try:
                self._generator.close()
            except Exception:
                pass
        self._generator = None

    def get(self) -> Optional[CoordinateGenerator]:
        return self._generator



import os
import re
import warnings
from dataclasses import dataclass
from typing import IO, Any, Callable, Dict, Generic, Tuple, TypeVar, Union

from tomlkit.container import Container
from tomlkit.items import Item

from ..exceptions import NECSTAccessibilityWarning, NECSTParameterNameError
from ..files import toml
from ..formatting.dict_to_html import html_repr_of_dict
from .parameters import Parameters

T_value = TypeVar("T_value")


@dataclass
class _RichParameter(Generic[T_value]):
    key: str
    value: T_value

    @property
    def parsed(self) -> Any:
        """Value which may be converted by attached parser."""
        if not hasattr(self, "_parsed_value"):
            _value = self.value
            if isinstance(_value, (Container, Item)):
                # TOML items will pass `isinstance()` checks but will fail `type()`
                # checks. Some scripts checks the type by `type()`, so we need to
                # convert them to corresponding Python built-in types.
                _value = _value.unwrap()
            self._parsed_value = self.parser(_value)
        return self._parsed_value

    @property
    def parser(self) -> Callable[[T_value], Any]:
        """Parser to convert value to another type or modify the value."""
        return self._parser if hasattr(self, "_parser") else lambda x: x

    @parser.setter
    def parser(self, __parser: Callable[[T_value], Any]) -> None:
        if hasattr(self, "_parsed_value"):
            # ``parsed`` property reuses the parsed value, so when the parser possibly
            # changed, the parsed value should be removed.
            del self._parsed_value
        self._parser = __parser

    def __str__(self) -> str:
        return str(self.parsed)


class RichParameters(Parameters):
    """Parameters with flexible look-up and high-level representation support.

    Parameters are not always just a key-value pair. Sometimes, we need higher level
    representation; all-in-one object that has rich information and methods about the
    parameter. This class provides a way to store such parameters.

    Parameters
    ----------
    __prefix
        Internally used prefix, which enables name-based filtering of parameters.
    **kwargs
        Parameters to be stored.

    Examples
    --------
    >>> params = RichParameters(a=1, b=2, **{"c[deg]": 3})

    You can attach a parser to a parameter, which will be used to convert the value to
    another high-level representation or modify the value.

    >>> params.attach_parser(b=lambda x: x * 2 * u.m)
    >>> params.b
    <Quantity 4. m>

    You can filter parameters by name, using the parameter prefix to accessing the
    attribute.

    >>> params = RichParameters(
    ...     general_param1=1,
    ...     specific_param1=2,
    ...     **{"general_param2[deg]": 3}
    ... )
    >>> params.general
    RichParameters
    general_param1 = 1 (_RichParameter)
    general_param2 = 3d00m00s (_RichParameter)

    """

    __slots__ = ("_prefix",)

    # Dot is used as namespace separator, so it's valid letter for parameter with
    # flexible look-up support
    _unit_matcher = re.compile(r"([\w\.]*)\[([\w/\s\*\^-]*)\]")

    def __init__(self, __prefix: str = "", /, **kwargs: Any) -> None:
        self._prefix = __prefix
        params = {
            k: v if isinstance(v, _RichParameter) else _RichParameter(k, v)
            for k, v in kwargs.items()
        }
        with warnings.catch_warnings():
            # The use of dots are valid in this class
            warnings.simplefilter("ignore", category=NECSTAccessibilityWarning)
            super().__init__(**params)
        self._parameters: Dict[str, _RichParameter[Any]]

        # Default aliases, which enable attribute access
        aliases = {k.replace(".", "_"): k for k in self._parameters.keys() if "." in k}
        self.attach_aliases(**aliases)

    @classmethod
    def from_file(cls, file: Union[os.PathLike, str, IO], /):
        """Read parameters from a TOML file.

        Parameters
        ----------
        file
            The file path or file object.

        Examples
        --------
        >>> params = RichParameters.from_file("path/to/file.toml")

        You can also read parameters from a file-like object.

        >>> with open("path/to/file.toml") as f:
        ...     params = RichParameters.from_file(f)

        And if you wish, remote files can be read:

        >>> params = RichParameters.from_file("https://example.com/file.toml")

        """
        file_path = isinstance(file, (os.PathLike, str))
        _params = toml.read(file)
        params = toml.flatten(_params)

        inst = cls(**params)
        if file_path:
            inst._metadata["path"] = file
        return inst

    def _parse(self, k: str, v: _RichParameter) -> Tuple[str, _RichParameter]:
        parsed_k, parsed_v = super()._parse(k, v.value)
        v.key = parsed_k
        v.value = parsed_v
        return parsed_k, v

    def attach_parsers(self, **kwargs: Callable[[Any], Any]) -> None:
        """Attach parsers to parameters.

        Parameters
        ----------
        **kwargs
            Parsers to be attached to parameters.

        Raises
        ------
        NECSTParameterNameError
            No parameter with the given name (key of the ``kwargs``) exists.

        Examples
        --------
        >>> params = RichParameters(a=1, b=2, **{"c[deg]": 3})
        >>> params.attach_parsers(b=lambda x: x * 2 * u.m)
        >>> params.b
        <Quantity 4. m>

        """
        for k, v in kwargs.items():
            if k in self._parameters:
                self._parameters[k].parser = v
            elif k in self._aliases:
                self._parameters[self._aliases[k]].parser = v

            try:
                self._pick(k).parser = v
            except KeyError:
                raise NECSTParameterNameError(f"Unknown parameter: {k!r}")

    def get(self, key: str, /) -> Any:
        """Flexibly look-up the parameters.

        Implementation of the ``__getitem__`` and ``__getattr__`` special methods.

        Parameters
        ----------
        key
            The name of the parameter.

        """
        try:
            return self._pick(key).parsed
        except KeyError:
            filtered = self._filter(key)

        if filtered:
            prefix = self._prefix + "_" + key if self._prefix else key
            inst = self.__class__(prefix, **filtered)
            inst._metadata = self._metadata
            for k, v in self._aliases.items():
                if v in filtered:
                    inst._aliases[k] = v
            return inst
        raise KeyError(key)

    def __getitem__(self, key: str, /) -> Any:
        return self.get(key)

    def __getattr__(self, key: str, /) -> Any:
        try:
            return self.get(key)
        except KeyError as e:
            # Should raise AttributeError to deny permissiveness to nonexistent
            # attributes. If an object performs as if it has any attributes by not
            # raising AttributeError, Jupyter Notebook won't trust the result of
            # `hasattr(obj, "_repr_html_")` hence don't display rich representations.
            # See also: https://github.com/jupyter/notebook/issues/2014
            raise AttributeError(f"No attribute {key!r}") from e

    def _filter(self, key: str, /) -> Dict[str, _RichParameter]:
        """Extract parameters which key starts with given key."""
        extracted = {}
        qualkey = self._prefix + "_" + key if self._prefix else key
        for k, v in self._parameters.items():
            if self._match(v.key, qualkey):
                extracted[k] = v
        for k, v in self._aliases.items():
            if self._match(k, qualkey):
                extracted[v] = self._parameters[v]
        return extracted

    def _pick(self, key: str, /) -> _RichParameter:
        """Pick-up at most one parameter, which key exactly matches to given one."""
        qualkey = self._prefix + "_" + key if self._prefix else key
        for v in self._parameters.values():
            if self._match(qualkey, v.key, strict=True):
                return v
        for k, v in self._aliases.items():
            if self._match(qualkey, k, strict=True):
                return self._parameters[v]
        raise KeyError(key)

    def _normalize(self, qualkey: str, /) -> str:
        """Normalize the key to enable structure insensitive access.

        Here normalization means replacing ``.`` with ``_``. This will convert keys to
        names accessible as attributes.

        Parameters
        ----------
        qualkey
            The key to be normalized. It should be a qualified key, i.e. the key with
            the prefix.

        """
        return qualkey.replace(".", "_")

    def _match(self, k1: str, k2: str, /, *, strict: bool = False) -> bool:
        """Judge whether namespace ``k1`` is equal to or contains key ``k2``.

        Parameters
        ----------
        k1
            The namespace. Must be a qualified key.
        k2
            The key. Must be a qualified key.
        strict
            If ``True``, this method judges ``k1`` and ``k2`` are equal or not.
            Otherwise, ``k1`` contains ``k2`` or not.

        """
        if strict:
            return self._normalize(k1) == self._normalize(k2)
        return self._normalize(k1).startswith(self._normalize(k2))

    @property
    def parameters(self) -> Dict[str, Any]:
        """Return a copy of the raw parameters."""
        return {k: v.parsed for k, v in self._parameters.items()}

    def _repr_html_(self) -> str:
        """Rich representation for Jupyter Notebook."""
        return html_repr_of_dict(
            {k: v.parsed for k, v in self._parameters.items()},
            type(self),
            aliases=self._aliases,
            metadata=dict(**self._metadata, prefix=self._prefix),
        )
